import datetime
import sys

from .typing import MetaModels, Union
from . import Providers
from . import Loggers


class Models(metaclass=MetaModels):
    
    class model:
        name: str
        base_provider: str
        best_site: str
    
    class gpt_35_turbo:
        name: str = 'gpt-3.5-turbo'
        base_provider: str = 'openai'
        best_site: str = Providers.Ails

    class gpt_4:
        name: str = 'gpt-4'
        base_provider: str = 'openai'
        best_site: str = Providers.Phind
        
class Utils:
    convert: dict = {
        'gpt-3.5-turbo': Models.gpt_35_turbo,
        'gpt-4': Models.gpt_4
    }

class ChatCompletion:
    @staticmethod
    def create(model: Models.model or str, messages: list, provider: Providers.Provider = None, log: Loggers.Logger = Loggers.Arweave, stream: bool = False, **kwargs):
        try:
            if isinstance(model, str):
                model = Utils.convert[model]
            
            engine = model.best_site if not provider else provider

            log_completion = _LogCompletion(log, messages, metadata=dict(model=model.name, engine=engine.__name__, kwargs=kwargs))
            output = engine._create_completion(model.name, messages, **kwargs)

            return (log_completion.stream(output)
                    if stream else log_completion.complete(output))

        except TypeError as e:
            print(e)
            arg: str = str(e).split("'")[1]
            print(
                f"ValueError: {engine.__name__} does not support '{arg}' argument", file=sys.stderr)
            sys.exit(1)

class _LogCompletion:
    def __init__(self, logger, input, metadata):
        self.logger = logger
        self.input = input
        self.metadata = metadata
        self.metadata['timestamp'] = datetime.datetime.now().isoformat()
    def stream(self, output):
        with self:
            for token in output:
                self.output += token
                yield token
    def complete(self, output):
        with self:
            self.output = ''.join(output)
            return self.output
    def __enter__(self):
        self.output = ''
        return self
    def __exit__(self, *exc):
        if self.logger is not None:
            locator = self.logger._log_completion(self.input, self.output, self.metadata)
            print(f"Logged {locator}", file=sys.stderr)
